/**
 ** com/charliemouse/cambozola/shared/CamStream.java
 **  Copyright (C) Andy Wilcock, 2001.
 **  Available from http://www.charliemouse.com
 **
 ** This file is part of the Cambozola package (c) Andy Wilcock, 2001.
 ** Available from http://www.charliemouse.com
 **
 **  Cambozola is free software; you can redistribute it and/or modify
 **  it under the terms of the GNU General Public License as published by
 **  the Free Software Foundation; either version 2 of the License, or
 **  (at your option) any later version.
 **
 **  Cambozola is distributed in the hope that it will be useful,
 **  but WITHOUT ANY WARRANTY; without even the implied warranty of
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 **  GNU General Public License for more details.
 **
 **  You should have received a copy of the GNU General Public License
 **  along with Cambozola; if not, write to the Free Software
 **  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **/
package com.charliemouse.cambozola.shared;

import java.awt.Image;
import java.awt.Toolkit;
import java.awt.image.ImageObserver;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.net.SocketException;
import java.net.URL;
import java.net.URLConnection;
import java.net.Socket;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public class CamStream extends Thread {
    public static final int CONNECT_STYLE_SOCKET = 1;
    public static final int CONNECT_STYLE_HTTP   = 2;


	private static final int IMG_FLUFF_FACTOR = 1;

    private ExceptionReporter m_reporter = null;
	private Vector m_listeners;
	private Object m_lock = new Object();
	private Image m_imain = null;
	private Toolkit m_tk;
	private URL m_stream;
	private URL m_docBase;
	private DataInputStream m_inputStream = null;
	private boolean m_isDefunct = false;
	private boolean m_collecting = false;
	private byte[] m_rawImage;
	private String m_imageType = "image/jpeg";
	private long m_startTime = 0;
	private int m_imgidx = 0;
	private int m_retryCount = 1;
	private int m_retryDelay = 1000;
	private String m_appName = "";
    private boolean m_debug = false;
    private int m_connectionMethod = CONNECT_STYLE_HTTP;

	public CamStream(URL strm, String app, URL docBase, int retryCount, int retryDelay, ExceptionReporter reporter, boolean debug)
	{
		m_tk = Toolkit.getDefaultToolkit();
		m_listeners = new Vector();
		m_stream = strm;
		m_appName = app;
        m_reporter = reporter;
		m_isDefunct = false;
		m_docBase = docBase;
		m_retryCount = retryCount;
		m_retryDelay = retryDelay;
        m_debug = debug;
	}


	public Image getCurrent()
	{
		synchronized (m_lock) {
			return m_imain;
		}
	}


	public byte[] getRawImage()
	{
		synchronized (m_lock) {
			return m_rawImage;
		}
	}


	public int getIndex()
	{
		synchronized (m_lock) {
			return m_imgidx;
		}
	}


	public String getType()
	{
		synchronized (m_lock) {
			return m_imageType;
		}
	}


	public URL getStreamURL()
	{
		return m_stream;
	}


	public double getFPS()
	{
		if (m_startTime == 0) {
			return 0.0;
		}
		long currTime = System.currentTimeMillis();
		return (1000.0 * (m_imgidx - IMG_FLUFF_FACTOR)) / (currTime - m_startTime);
	}


	public void addImageChangeListener(ImageChangeListener cl)
	{
		m_listeners.addElement(cl);
	}


	public void removeImageChangeListener(ImageChangeListener cl)
	{
		m_listeners.removeElement(cl);
	}


	private void fireImageChange()
	{
        if (m_debug) {
            System.err.println("// " + System.currentTimeMillis() + " - Notification of image update sent");
        }
		ImageChangeEvent ce = new ImageChangeEvent(this);
		for (Enumeration e = m_listeners.elements(); e.hasMoreElements();) {
			((ImageChangeListener) e.nextElement()).imageChanged(ce);
		}
	}


	public void run()
	{
		StreamSplit ssplit = null;
		try {
            String encpasswd = null;
			//
			// Loop for a while until we either give up (hit m_retryCount), or
			// get a connection.... Sleep inbetween.
			//
			String connectionError = null;
			String ctype = null;
			Hashtable headers = null;
			int tryIndex = 0;
			int retryCount = m_retryCount;
			int retryDelay = m_retryDelay;
			//
			do {
				//
				// Keep track of how many times we tried.
				//
				tryIndex++;

                if (m_debug)
                {
                    System.err.println("// Connection type = " + m_connectionMethod + ", URL = " + m_stream);
                }
                if (m_connectionMethod == CONNECT_STYLE_SOCKET)
                {
                    //
                    // Original [DIRECT] method.
                    //
                    int port = m_stream.getPort();
                    Socket websock = new Socket(m_stream.getHost(), port);
                    DataOutputStream outputStream = new DataOutputStream(websock.getOutputStream());
                    m_inputStream = new DataInputStream(new BufferedInputStream(websock.getInputStream()));
                    StringBuffer request = new StringBuffer();
                    request.append("GET " + m_stream.getFile() + " HTTP/1.0\r\n");
                    if (m_docBase != null) {
                        request.append("Referer: " + m_docBase + "\r\n");
                    }
                    request.append("User-Agent: " + m_appName + "\r\n");
                    if (encpasswd != null) {
                        request.append("Authorization: Basic " + encpasswd + "\r\n");
                    }
                    request.append("Host: " + m_stream.getHost() + "\r\n");
                    request.append("\r\n");
                    outputStream.writeBytes(request.toString());
                    outputStream.close();
                    //
                    ssplit = new StreamSplit(m_inputStream);
                    headers = ssplit.readHeaders();
                } else {
                    //
                    // Better method - access via URL Connection
                    //
                    URLConnection conn = m_stream.openConnection();
                    if (m_docBase != null) {
                        conn.setRequestProperty("Referer", m_docBase.toString());
                    }
                    conn.setRequestProperty("User-Agent", m_appName);
                    conn.setRequestProperty("Host", m_stream.getHost());
                    m_inputStream = new DataInputStream(new BufferedInputStream(conn.getInputStream()));
                    //
                    // Read Headers for the main thing...
                    //
                    ssplit = new StreamSplit(m_inputStream);
                    headers = ssplit.readHeaders(conn);
                }
                if (m_debug)
                {
                    System.err.println("// Request sent; Main Response headers:");
                    for (Enumeration enm = headers.keys();enm.hasMoreElements();)
                    {
                        String hkey = (String)enm.nextElement();
                        System.err.println("//   " + hkey + " = " + headers.get(hkey));
                    }
                }
                //
				m_collecting = true;
				//
				// Work out the content type/boundary.
				//
				connectionError = null;
				ctype = (String) headers.get("content-type");
				if (ctype == null) {
					connectionError = "No main content type";
				} else if (ctype.indexOf("text") != -1) {
					String response = null;
					while ((response = m_inputStream.readLine()) != null) {
						System.out.println(response);
					}
					connectionError = "Failed to connect to server (denied?)";
				}
				if (connectionError == null)
				{
					break; // Yay!! got one.
				} else if (m_isDefunct) {
					//
					// Not wanted any more...
					//
					return;
				} else {
					//
					// Wait a while before retrying...
					//
                    if (m_debug)
                    {
                        System.err.println("// Waiting for " + retryDelay + " ms");
                    }
					m_reporter.reportFailure(connectionError);
					sleep(retryDelay);
				}
			} while (tryIndex < retryCount);
			//
			if (connectionError != null)
			{
				return;
			}
			//
			// Boundary will always be something - '--' or '--foobar'
            //
			int bidx = ctype.indexOf("boundary=");
			String boundary = StreamSplit.BOUNDARY_MARKER_PREFIX;
			if (bidx != -1) {
				boundary = ctype.substring(bidx + 9);
				ctype = ctype.substring(0, bidx);
				if (!boundary.startsWith(StreamSplit.BOUNDARY_MARKER_PREFIX)) {
					boundary = StreamSplit.BOUNDARY_MARKER_PREFIX + boundary;
				}
			}
			//
			// Now if we have a boundary, read up to that.
			//
			if (ctype.startsWith("multipart/x-mixed-replace")) {
                if (m_debug)
                {
                    System.err.println("// Reading up to boundary");
                }
				ssplit.skipToBoundary(boundary);
			}

			do {
				if (m_collecting) {
					//
					// Now we have the real type...
					//  More headers (for the part), then the object...
					//
					if (boundary != null) {
						headers = ssplit.readHeaders();
                        if (m_debug) {
                            System.err.println("// Chunk Headers recieved:");
                            for (Enumeration enm = headers.keys();enm.hasMoreElements();)
                            {
                                String hkey = (String)enm.nextElement();
                                System.err.println("//   " + hkey + " = " + headers.get(hkey));
                            }
                        }
						//
						// Are we at the end of the m_stream?
						//
						if (ssplit.isAtStreamEnd()) {
							break;
						}
						ctype = (String) headers.get("content-type");
						if (ctype == null) {
							throw new Exception("No part content type");
						}
					}
					//
					// Mixed Type -> just skip...
					//
					if (ctype.startsWith("multipart/x-mixed-replace")) {
						//
						// Skip
						//
						bidx = ctype.indexOf("boundary=");
						boundary = ctype.substring(bidx + 9);
						//
                        if (m_debug) {
                            System.err.println("// Skipping to boundary");
                        }
						ssplit.skipToBoundary(boundary);
					} else {
						//
						// Something we want to keep...
						//
                        if (m_debug) {
                            System.err.println("// Reading to boundary");
                        }
						byte[] img = ssplit.readToBoundary(boundary);
						if (img.length == 0) {
							break;
						}
						//
						// FPS counter.
						//
						if (m_imgidx > IMG_FLUFF_FACTOR && m_startTime == 0) {
							m_startTime = System.currentTimeMillis();
						}
						//
						// Update the image [fores events off]
						//
						updateImage(ctype, img);
					}
				}
				try {
					Thread.sleep(20);
				} catch (InterruptedException ie) {
				}
			} while (!m_isDefunct);
		} catch (Exception e) {
            if (!m_collecting)
            {
                m_reporter.reportFailure(e.toString());
            } else if (!m_isDefunct) {
				m_reporter.reportError(e);
			}
		} finally {
			unhook();
		}
		//
		// At this point, the m_stream m_inputStream done
		// [could dispplay a that's all folks - leaving it as it m_inputStream
		//  will leave the last frame up]
		//
	}

	private void updateImage(String ctype, byte[] img)
	{
		//
		// Update our image...
		//
		synchronized (m_lock) {
			m_imageType = ctype;
			m_imain = m_tk.createImage(img);
			m_rawImage = img;
			m_imgidx++;

            m_imain.getWidth(new ImageObserver()
            {
                public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height)
                {
                    if (m_debug) {
                        System.err.println("// Image Progress = " + infoflags);
                    }
                    boolean fully = ((infoflags & (ImageObserver.ALLBITS | ImageObserver.PROPERTIES)) != 0);
                    if (fully) {
                        fireImageChange();
                    }
                    return !fully;
                }
            });
        }
	}

	public void finalize()
	{
		unhook();
	}


	public void unhook()
	{
		m_collecting = false;
		m_isDefunct = true;
		try {
			if (m_inputStream != null)
			{
				m_inputStream.close();
			}
			m_inputStream = null;
		} catch (Exception e) {
		}
	}
}